
import de.mud.jta.Plugin;
import de.mud.jta.FilterPlugin;
import de.mud.jta.PluginBus;
import de.mud.jta.PluginConfig;
import de.mud.jta.event.OnlineStatusListener;
import de.mud.jta.event.ConfigurationListener;

import java.awt.Window;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import java.util.Date;
import java.util.EventObject;

//import javax.comm.CommPort;
//import javax.comm.SerialPort;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JToggleButton;

import com.lucent.kermit.Kermit;
import com.lucent.kermit.KermitController;

/**
 * This file implements a filter for kermit
 */

public class FileTransfer extends Plugin implements
  FilterPlugin, OnlineStatusListener, ActionListener, KermitController {

  private PluginConfig		config;
  protected FilterPlugin	source;
  protected boolean		onlineStatus = false;

  protected Kermit	kermit = null;		// kermit object
  protected byte[]	lastKermitPkt = null;	// last packet we sent
  protected File	kermitFile = null;	// local file we have open
  protected File	kermitDir = null;	// Dir used in file dialog
  protected int		kermitSeq;		/* sending: the next number we
						 * are going to send.
						 * receiving: the seqNo that
						 * came in the last good packet
						 */

  protected JButton	kermitCancelButton;
  protected boolean	parity = false;		// Use binary prefix if true

  ProgressFrame progressFrame;	// progress dialog
  static final protected int		debug = 0;
  static final protected byte[]		punctuationBytes
    = new String( "!@#$%^&*()_+-=[]{}\\|;:'\",.<>/?~" ).getBytes();
  
  /**
   * @param bus is the JTA filter bus
   * @param id is the name by which this plugin is referred
   *
   * @see PluginBus
   */
  public FileTransfer( final PluginBus bus, final String id ) {
    super( bus, id );
    bus.registerPluginListener( this );
    bus.registerPluginListener( new ConfigurationListener() {
	public void setConfiguration(PluginConfig newConfig) {
	  dbg( 1, "config:" + newConfig );
	  config = newConfig;
	}
      });

    kermitDir = new File( "." );
  }

  public void online() {
    onlineStatus = true;
    /*
    TelnetPanel telnetPanel = TacTool.getTelnetPanel();
    if( TacTool.getTelnetPanel().serialPortPlugin != null ) {
      CommPort port = TacTool.getTelnetPanel().serialPortPlugin.port;
      if( port instanceof javax.comm.SerialPort ) {
	if( ((SerialPort)port).getParity() != SerialPort.PARITY_NONE ) {
	  parity = true;
	}
      }
    }
    */
  }

  public void offline() {
    onlineStatus = false;
    shutdown();
    parity = false;
  }

  public void showKermitStatus() {
    progressFrame = new ProgressFrame( Main.frame );
    progressFrame.setTitle( "Kermit progress" );
    /*
    progressFrame.
      setConnectionName( TacTool.getTelnetPanel().lastConnectedHost );
    progressFrame.setLocation(
		      TacUtil.getCenteredPoint( TacTool.getTacToolParent(),
						progressFrame ) );
    */
    kermitCancelButton = progressFrame.getCancelButton();
    kermitCancelButton.addActionListener( this );
    progressFrame.setVisible( true );
  }

  public void shutdown() {
    if( kermit != null ) kermit.shutdown();
    kermit = null;
    if( progressFrame != null ) {
      progressFrame.setCompleted();
      progressFrame = null;
    }
  }

  protected int getPropInt( String prop, int defInt ) {
    String s = config.getProperty( "FileTransfer", id, prop );
    if( s == null ) return defInt;
    try {
      return Integer.parseInt( s );
    } catch( Exception e ) {
      return defInt;
    }
  }

  public void startKermit() {
    kermit.setDefaultMaxPacketLength( getPropInt("maxPacketSize",1024));
    kermit.setDefaultSOH( getPropInt( "blockStart", 1 ) );
    kermit.setDefaultEndOfLine( getPropInt( "endOfLine", 13 ) );
    kermit.setDefaultControlPrefix( getPropInt("controlPrefix", '#' ));
    kermit.setDefaultBinaryPrefix( getPropInt( "binaryPrefix", '&' ) );
    kermit.setDefaultCheckSum( getPropInt( "checksum", 1 ) );
    kermit = new Kermit( this );
    if( parity ) kermit.setParity( 1 );
  }

  public void actionPerformed( ActionEvent e ) {
    if( e.getActionCommand() == "Kermit Receive" ) {
      // First pick a directory
      boolean done = false;
      File dir = null;

      while( !done ) {
	dir = browseForFiles( kermitDir, JFileChooser.DIRECTORIES_ONLY,
			      "Select Output Directory" );

	if( dir == null ) {
	  startKermit();
	  kermit.sendError( "Cancelled By User" );
	  kermit = null;
	  return;
	}

	if( !( dir.exists() && dir.isDirectory() ) ) {
	  JOptionPane.showMessageDialog( null,
					 "Directory "+dir.getAbsolutePath()
					 +" does not exist" );
	  continue;
	}
	done = true;
      }

      startKermit();
      kermitDir = dir;

      /* At this stage, the remote side has been told to send a file.  It will
       * pause for a few seconds and then start sending Send-init pkts.
       * Until we turn the kermit flag on by selecting an directory name
       * where we will place all the input files.  Until then, these
       * send-init packets will show up on the terminal.
       *
       * This should be harmless.
       */

      showKermitStatus();
      progressFrame.setTitle( "Receiving file from remote system" );
      progressFrame.setStatus( "Waiting for initialization..." );
      statusChanged( kermit );
    }

    else if( e.getActionCommand() == "Kermit Send" ) {
      /* At this stage, the remote side has been told to receive a file.  It
       * will pause for a few seconds and then start sending NAK packets. Until
       * we turn the kermit flag on by selecting a local output filename, these
       * send-init packets will show up on the terminal.
       *
       * This should be harmless.
       */

      // start kermit
      // get the filename and then kermit.file( file )
      // let kermit open the file

      setInputFile();

      startKermit();
      if( kermitFile == null ) {
	kermit.sendError( "Cancelled by user" );
	kermit = null;
	return;
      }

      showKermitStatus();
      progressFrame.setTitle( "Sending file to remote system" );
      progressFrame.setStatus( "Waiting for attributes..." );

      kermit.sendFile( kermitFile );
      statusChanged( kermit );
    }

    else if( e.getSource() == kermitCancelButton ) {
      sendError( "Cancelled by user" );
    }

    else {
      dbg( 0, e.getActionCommand() );
    }
  }

  public void setFilterSource(FilterPlugin source) {
    dbg( 1, "Connected to source:" + source );
    this.source = source;
  }

  /* Read from the remote end of the pipeline.  Print to stderr
   * the eavesdropping object.
   */
  public int read(byte[] b) throws IOException { // pass everything through
    // If kermit is on, we need to timeout this upcoming read.
    int n;
    if( kermit == null ) {
      n = source.read(b);
    } else {
      try {
	n = (new TimedPluginRead(source, kermit.getTimeout()*1000 )).read( b );
      } catch( InterruptedException ex ) {

	// The thread was interrupted. - kermit may have gone away on this
	// last read, so....
	if( kermit != null ) kermit.timeout();
	n = 0; // return null on all interrupted packets
      }
    }

    //    if( kermit == OFF ) dumpBuffer( "<-", b, n );

    // don't process empty reads
    if( n <= 0 ) {
      //      System.err.println( "<-(null)" );
      return n;
    }

    if( kermit != null ) {
      if( debug > 0 ) dumpBuffer( "kermit <-", b, n );
      kermit.kermitGotPacket( b, n );
      statusChanged( kermit );
      return 0;
    }

    return n;
  }

  void sendError( String error ) {
    if( kermit != null ) kermit.sendError( error );
    if( progressFrame != null )
      progressFrame.setStatus( error );
    else
      JOptionPane.showMessageDialog( null, error );
    shutdown();
  }

  /** Write to the pipeline sink
   */
  public void write(byte[] b) throws IOException { // pass everything through
    if( source == null ) return;
    source.write( b );
  }

  protected void setInputFile() {
    kermitFile = browseForFiles( kermitDir, JFileChooser.FILES_ONLY,
				 "Select File to Send" );
    if( kermitFile == null ) return;

    kermitDir = kermitFile.getParentFile();
  }

  protected class TimedPluginRead implements Runnable {
    int			timeoutMillis = 10000; // 10 second default
    FilterPlugin	plugin;
    byte[]		byteArray;
    int			bytesRead = 0;
    IOException		ioException = null;
    Thread		thread = null;

    TimedPluginRead( FilterPlugin filterPlugin, int millis ) {
      timeoutMillis = millis;
      plugin = filterPlugin;
    }

    public void run() {
      try {
	bytesRead = source.read( byteArray );
      } catch( IOException ex ) {
	ioException = ex;
      }
    }

    int read( byte[] b ) throws IOException, InterruptedException {
      byteArray = b;
      Thread thread = new Thread( this, "TimedPluginRead" );
      thread.start();
      thread.join( timeoutMillis );
      if( ioException != null ) throw ioException;
      if( thread.isAlive() ) { // we are still waiting for input.
	// Kill this sucker!
	thread.interrupt();
	thread = null;
	throw new InterruptedException( "timed out waiting for read" );
      }
      return bytesRead;
    }
  }

  public File kcGetFileToReceive( Kermit kermit ) {
    progressFrame.setFromFile( kermit.getRemoteFileName() );
    progressFrame.setStatus( "Opening local file..." );
    File kFile = new File( kermit.getRemoteFileName() );
    if( kFile.exists() ) {
      int n = JOptionPane.showConfirmDialog( null,
					     "Do you want to overwrite:"
					     + kFile.getName(),
					     "Overwrite full",
					     JOptionPane.YES_NO_OPTION );

      if( n != JOptionPane.YES_OPTION ) {
	progressFrame.toFront();
	progressFrame.setValue( 0 );
	sendError( "Cancelled by user" );
	return null;
      }
    }

    progressFrame.setToFile( kFile.getAbsolutePath() );
    progressFrame.toFront();
    return kFile;
  }

  protected void statusChanged( Kermit kermit ) {
    if( progressFrame != null ) {
      updateProgressFrame( kermit );
    }
    //    dbg( 0, kermit.toString() );
    // should always be last since shutdown removes kermit and progressFrame
    if( kermit == null )
      shutdown();
    else
      if( kermit.getState() == kermit.OFF ) shutdown();
  }

  protected void updateProgressFrame( Kermit kermit ) {
    if( progressFrame == null ) return;

    if( kermit.getState() == kermit.FH_PENDING ||
	kermit.getState() == kermit.RECV_DATA_PENDING ) {
      // We are sending or receiving a new file header.  Reset progress
      progressFrame.setMaximum( 1 );
      progressFrame.setValue( 0 );
    }

    if( kermit.getState() == kermit.DATA_PENDING &&
	kermit.getLocalFile() != null && kermit.getByteProgress() <= 0 ) {
      // we have a local file, but we haven't started yet.  This is a good
      // time to set the file names on the send side
      progressFrame.setToFile( kermit.getRemoteFileName() );
      progressFrame.setFromFile( kermit.getLocalFile().getAbsolutePath() );
    }

    progressFrame.setStatus( kermit.getStatus() );
    if( progressFrame.getMaximum() <= 1 ) { // lowest value
      progressFrame.setMaximum( kermit.getFileSize() );
    }
    progressFrame.setValue( kermit.getByteProgress() );
  }

  public void kcWrite( Kermit kermit, byte[] packet ) {
    if( debug > 0 ) dumpBuffer( "-> kermit", packet, packet.length );
    try {
      write( packet );
    } catch( Exception e ) {
      shutdown();
      sendError( "Could not write to socket:" + e );
    }
  }

  static protected File browseForFiles( File dir, int mode, String title ) {
    JFileChooser fc = new JFileChooser();
    fc.setApproveButtonText( "Select" );
    if( dir != null ) fc.setCurrentDirectory( dir );
    fc.setFileSelectionMode( mode );
    int returnVal = fc.showDialog( null, title );
    if( returnVal == JFileChooser.APPROVE_OPTION )
      return fc.getSelectedFile();
    return null;
  }

  private void dbg( int _debug, Object msg ) {
    if( debug >= _debug )
      System.err.println( getClass().getName() + ":" + msg );
  }

  static void dumpBuffer( String label, byte[] byteArray, int length ) {
    System.err.println( label + ":" + length );
    int index = 0;
    int headingIndex = 10;
    while( index < length ) {
      if( headingIndex++ >= 10 ) {
	headingIndex = 1;
	for( int i = 0; i < 16 && index < length; i++ ) {
	  String s = ""+i;
	  if( s.length() == 1 ) s = "0"+s;
	  System.err.print( " 0" + s );
	}
	System.err.println("");
      }

      for( int i = 0; i < 16 && index < length; i++ ) {
	byte b = byteArray[index++];
	System.err.print( " " + printChar( b ) );
      }
      System.err.println("");
    }
  }
  static private String printChar( byte b ) {
    char c = (char) b;

    // check if the byte is punctuation
    if( c < 0177 ) {
      if( Character.isDigit( c )
	  || Character.isLetter( c )
	  || Character.isSpaceChar( c )
	  ) {
	return "  " + c;
      }
    }

    for( int j = 0; j < punctuationBytes.length; j++ ) {
      if( b == punctuationBytes[j] ) return "  "+c;
    }

    // The byte is a 'non-printable character'
    int iByte = b & 0377;
    String s = Integer.toOctalString( iByte );
    if( s.length() == 3 ) return s;
    if( s.length() == 2 ) return "0" + s;
    return "00" + s;
  }
}
